__all__ = 'EnvoyProject', 'EnvoyFile', 'GridMarketsAPIClient'

import pathlib

from gm_meta_plugin.gridmarkets.remote_project import convert_packed_project
from gm_meta_plugin.gridmarkets.xml_api_schema_parser import XMLAPISchemaParser
from . import constants as api_constants
from gm_meta_plugin.api_client import APIClient as MetaAPIClient
from gm_meta_plugin.cached_value import CachedValue
from gm_meta_plugin.user_info import UserInfo
from gm_meta_plugin.machine_option import MachineOption
from gm_meta_plugin.product import Product
from gm_meta_plugin.utils import get_files_in_directory, get_exception_with_traceback, get_desktop_path
from gm_meta_plugin.errors import *
from .. import attribute_types

import gridmarkets
from gridmarkets.errors import AuthenticationError, APIError, InvalidRequestError


class EnvoyProject:
    def __init__(self, name, deleting, num_files, total_size, auto_downloading,
                 watched_list):
        self._name = name
        self._deleting = deleting
        self._num_files = num_files
        self._total_size = total_size
        self._auto_downloading = auto_downloading
        self._watched_list = watched_list

    def get_name(self):
        return self._name

    def is_deleting(self):
        return self._deleting

    def get_file_count(self):
        return self._num_files

    def get_total_size(self):
        return self._total_size

    def is_auto_downloading(self):
        return self._auto_downloading

    def get_watched_list(self):
        return self._watched_list


class EnvoyFile:
    def __init__(self, check_sum, last_modified, name, key, size, fmt_name):
        self._check_sum = check_sum
        self._last_modified = last_modified
        self._name = name
        self._key = key
        self._size = size
        self._fmt_name = fmt_name

    def get_check_sum(self):
        return self._check_sum

    def get_last_modified(self):
        return self._last_modified

    def get_name(self):
        return self._name

    def get_key(self):
        return self._key

    def get_size(self):
        return self._size

    def get_fmt_name(self):
        return self._fmt_name


class GridMarketsAPIClient(MetaAPIClient):
    http_api_endpoint = u"https://api.gridmarkets.com:8003/api/render/1.0/"

    def __init__(self, plugin, logging_coordinator, factory_collection):
        MetaAPIClient.__init__(self, plugin)
        self._factory_collection = factory_collection
        self._signed_in = False
        self._envoy_client = None
        self._log = logging_coordinator.get_logger(u"GridMarketsAPIClient")

        # used for caching projects
        self._api_schema_cache = CachedValue(None)
        self._remote_projects_cache = CachedValue([])
        self._project_files_dictionary_cache = {}
        self._product_resolver = CachedValue(None)
        self._products_cache = CachedValue([])
        self._cached_machine_options = CachedValue({})
        self._cached_user_info = CachedValue(UserInfo(9999, 60))
        self._cached_available_credits = CachedValue(float(0))

    def clear_all_caches(self):
        self._log.info(u"Clearing cached API data")
        self._remote_projects_cache.reset_and_clear_cache()
        self._project_files_dictionary_cache.clear()
        self._product_resolver.reset_and_clear_cache()
        self._products_cache.reset_and_clear_cache()
        self._cached_machine_options.reset_and_clear_cache()
        self._cached_user_info.reset_and_clear_cache()
        self._cached_available_credits.reset_and_clear_cache()

    def sign_in(self, user, skip_validation = False):
        self._log.info(u"Signing in as " + user.get_auth_email())

        MetaAPIClient.sign_in(self, user, skip_validation=skip_validation)

        self._envoy_client = gridmarkets.EnvoyClient(email=user.get_auth_email(), access_key=user.get_auth_key())

        # refresh caches
        self.clear_all_caches()
        self.get_root_directories(ignore_cache=True)
        self.get_api_schema(ignore_cache=True)
        self.get_user_info(ignore_cache=True)
        self.get_available_credits(ignore_cache=True)
        self._signed_in = True

    def sign_out(self):
        if self.is_user_signed_in():
            self._log.info(u"Signing out user " + self.get_signed_in_user().get_auth_email())

        MetaAPIClient.sign_out(self)
        self._signed_in = False

    def is_user_signed_in(self):
        return self._signed_in

    def connected(self):
        raise NotImplementedError

    def get_signed_in_user(self):
        return MetaAPIClient.get_signed_in_user(self)

    def get_cached_api_schema(self):
        return self._api_schema_cache.get_value()

    def get_api_schema(self, factory_collection=None, ignore_cache=False):
        self._log.info(u"Getting api schema. ignore_cache=" + unicode(ignore_cache))

        if factory_collection is None:
            factory_collection = self._factory_collection

        if ignore_cache or not self._api_schema_cache.is_cached():
            self._log.info(u"Fetching api schema from XML source.")

            try:
                api_schema = XMLAPISchemaParser.parse(self.get_plugin(), factory_collection)
                api_schema.trim_unsupported_applications(self.get_plugin())
                self._api_schema_cache.set_value(api_schema)
            except Exception as e:
                msg = u"Could not fetch api schema." + get_exception_with_traceback(e)
                self._log.error(msg, exc_info=True)
                raise APIClientError(msg)

        self._log.info(u"Returning api schema")
        return self._api_schema_cache.get_value()

    @staticmethod
    def validate_credentials(auth_email, auth_access_key):

        # check an email address has been entered
        has_email = isinstance(auth_email, unicode) and len(auth_email) > 0

        # check a auth key has been entered
        has_key = isinstance(auth_access_key, unicode) and len(auth_access_key) > 0

        if has_email and has_key:
            try:
                client = gridmarkets.EnvoyClient(email=auth_email, access_key=auth_access_key)
                client.validate_auth()
            except AuthenticationError as e:
                raise InvalidUserError(message=e.user_message)
            except APIError as e:
                e_msg = unicode(e.user_message)

                # pretty error message for common errors
                if e_msg.startswith(u"Unexpected error communicating with Envoy.Please start Envoy"):
                    e_msg = u"Unable to connect to Envoy."

                raise APIClientError(message=e_msg)

        elif not (has_email or has_key):
            message = u"No " + api_constants.COMPANY_NAME + u" email address or access key provided."
            raise InvalidUserError(message=message)
        elif not has_email:
            message = u"No " + api_constants.COMPANY_NAME + u" email address provided."
            raise InvalidEmailError(message=message)
        elif not has_key:
            message = u"No " + api_constants.COMPANY_NAME + u" access key provided."
            raise InvalidAccessKeyError(message=message)

        return True

    def upload_project(self,
                       packed_project,
                       upload_root_dir,
                       delete_local_files_after_upload = False):

        if not self.is_user_signed_in():
            raise NotSignedInError(u"Must be signed-in to upload a project.")

        self._log.info(u"Uploading packed project")

        project_dir = unicode(packed_project.get_root_dir())
        project_name = packed_project.get_name()

        gm_project = gridmarkets.Project(project_dir, project_name)

        if upload_root_dir:
            # add all files to packed project files list
            files = get_files_in_directory(packed_project.get_root_dir())
            packed_project.add_files(files)

            self._log.info(u"Marking directory \"" + project_dir + u"\" for upload.")
            gm_project.add_folders(project_dir)

        else:
            # get project files as strings
            files = self._get_files_to_upload(packed_project)

            for file in files:
                self._log.info(u"Marking \"" + file + u"\" for upload.")

            gm_project.add_files(*files)

        try:
            self._envoy_client.upload_project_files(gm_project)
        except APIError as e:
            raise APIClientError(e.user_message)

        self._log.info(u"Uploaded project")

        return convert_packed_project(packed_project, self.get_plugin())


    def _get_gm_job(self, job_preset, remote_project):
        self._log.info(u"Calculating Job parameters...")
        job_settings = {}
        for job_preset_attribute in job_preset.get_job_preset_attributes():

            # skip attributes which are marked as skip
            if job_preset_attribute.get_job_attribute().get_skip():
                continue

            key = job_preset_attribute.get_key()

            # The python envoy client treats these attributes as something other than job parameters so we skip them
            # here and deal with them elsewhere.
            if key in [api_constants.API_KEYS.AUTO_DOWNLOAD,
                       api_constants.API_KEYS.REMOTE_OUTPUT_FOLDER_NAME,
                       api_constants.API_KEYS.LOCAL_DOWNLOAD_PATH]:
                continue

            value = job_preset_attribute.eval(self.get_plugin(), remote_project)
            job_settings[key] = value
            self._log.info(u"Job Parameter '" + key + u"': " + unicode(value))

            # The python envoy client treats machine type and gpu as two different attributes so we must infer
            # the gpu attribute from the machine type here.
            if key == api_constants.API_KEYS.MACHINE_TYPE:

                # we need to find the machine type with this id and work out if it is a gpu machine or not
                machine_id = value
                is_gpu = self._is_gpu_enabled_machine(machine_id)
                job_settings[api_constants.API_KEYS.GPU] = is_gpu
                self._log.info(u"Job Parameter '" + api_constants.API_KEYS.GPU + u"': " + unicode(is_gpu))

        job = gridmarkets.Job(**job_settings)

        # Check the app is supported
        if job.app not in self.get_product_app_types():
            msg = u"App \"" + job.app + u"\" is not supported for your account. Please contact " + \
                  api_constants.COMPANY_NAME + u" support for details."
            self._log.error(msg)
            raise UnsupportedApplicationError(msg)

        # Check the version is supported
        if job.app_version not in self.get_product_versions(job.app):
            msg = u"Version \"" + job.app_version + u"\" for app_type \"" + job.app + \
                  u"\" is not supported for your account. Please contact " + api_constants.COMPANY_NAME + \
                  u" support for details."

            self._log.error(msg)
            raise UnsupportedApplicationError(msg)

        return job

    def _get_remote_output_folder_name(self, job_preset, remote_project):
        attribute = job_preset.get_job_preset_attribute_by_key(api_constants.API_KEYS.REMOTE_OUTPUT_FOLDER_NAME)
        remote_output_folder_name = attribute.eval(self.get_plugin(), remote_project)
        return remote_output_folder_name

    def _get_local_download_path(self, job_preset, remote_project):
        attribute = job_preset.get_job_preset_attribute_by_key(api_constants.API_KEYS.LOCAL_DOWNLOAD_PATH)
        local_download_path = attribute.eval(self.get_plugin(), remote_project)
        return local_download_path

    def _get_skip_auto_download(self, job_preset, remote_project):
        attribute = job_preset.get_job_preset_attribute_by_key(api_constants.API_KEYS.AUTO_DOWNLOAD)
        auto_download = attribute.eval(self.get_plugin(), remote_project)
        return not auto_download

    def _set_watch_files(self, gm_project, job_preset, remote_project):

        timestamp = unicode(gm_project.timestamp).replace(' ', '_').replace(':', '-').replace('.', '_')

        # setup the download path
        download_path = self._get_local_download_path(job_preset, remote_project)
        if not download_path:
            download_path = "{0}/{1}/render_results/{2}".format(get_desktop_path(),
                                                                remote_project.get_name(),
                                                                timestamp)

        # setup the remote output folder pattern
        remote_output_folder_name = self._get_remote_output_folder_name(job_preset, remote_project)
        if not remote_output_folder_name:
            remote_output_folder_name = timestamp

        remote_output_folder = "{0}/render_results/{1}".format(gm_project.remote_root, remote_output_folder_name)
        download_pattern = "{0}/.+".format(remote_output_folder)

        # create the default watch file
        default_watch_file = gridmarkets.WatchFile(download_pattern, download_path)

        # add the watch files
        gm_project.add_watch_files(default_watch_file)

        gm_project.remote_output_folder_name = remote_output_folder_name
        gm_project.remote_output_folder = remote_output_folder

        for watch_file in gm_project.watch_files:
            self._log.info(u"Watch File \"" + unicode(watch_file.output_pattern) + u"\" : \"" +
                           unicode(watch_file.download_path) + u"\"")

    def _get_files_to_upload(self, packed_project):

        def encode_path(path):
            import chardet
            encoded_path = str(path).decode(chardet.detect(str(path))['encoding'])
            return encoded_path

        return list(map(lambda x: encode_path(x), list(packed_project.get_files())))

    def submit_new_project(self, packed_project, job_preset, delete_local_files_after_upload = False):

        if not self.is_user_signed_in():
            raise NotSignedInError(u"Must be signed-in to Submit a job.")

        self._log.info(u"Preparing to submit current scene as project \"" + packed_project.get_name() +
                       u"\" with job \"" + job_preset.get_name() + u"\".")

        client = self._envoy_client
        project_dir = str(packed_project.get_root_dir())
        project_name = packed_project.get_name()

        gm_project = gridmarkets.Project(project_dir, project_name)

        files_to_upload = self._get_files_to_upload(packed_project)
        gm_project.add_files(*files_to_upload)

        remote_project = convert_packed_project(packed_project, self.get_plugin())

        job = self._get_gm_job(job_preset, remote_project)

        # add job to project
        gm_project.add_jobs(job)

        # set the watch files
        self._set_watch_files(gm_project, job_preset, remote_project)

        # get the skip auto download flag
        skip_auto_download = self._get_skip_auto_download(job_preset, remote_project)

        # submit the job to the remote project
        self._log.info(u"Submitting current scene with job \"" + job_preset.get_name() + u"\" to project \"" +
                       project_name + u"\".")

        try:
            client.submit_project(gm_project, skip_upload=False, skip_auto_download=skip_auto_download)

        except InvalidRequestError as e:
            msg = e.user_message
            if e.errors:
                msg = msg + u"\n" + unicode(e.errors)

            raise APIClientError(msg)
        except APIError as e:
            raise APIClientError(e.user_message)

        self._log.info(u"Job submitted - See Render Manager for details")

        return remote_project

    def submit_to_remote_project(self, remote_project, job_preset):

        if not self.is_user_signed_in():
            raise NotSignedInError(u"Must be signed-in to Submit a job.")

        # check the project exists
        if len(remote_project.get_files(force_update=True)) <= 0:
            msg = u"Can not submit job to remote project \"" + remote_project.get_name() + \
                  u"\", project does not exist."
            self._log.warning(msg)
            raise APIClientError(msg)

        # check the project has finished uploading
        if remote_project.get_remaining_bytes_to_upload(force_update=True) != 0:
            msg = u"Can not submit job to remote project \"" + remote_project.get_name() + \
                  u"\", project has not finished uploading."
            self._log.warning(msg)
            raise APIClientError(msg)

        self._log.info(u"Preparing job " + job_preset.get_name() + u" for submission to remote project " +
                       remote_project.get_name())

        client = self._envoy_client
        gm_project = gridmarkets.Project(str(remote_project.get_root_dir()), remote_project.get_name())

        job = self._get_gm_job(job_preset, remote_project)

        # add job to project
        gm_project.add_jobs(job)

        # set the watch files
        self._set_watch_files(gm_project, job_preset, remote_project)

        # get the skip auto download flag
        skip_auto_download = self._get_skip_auto_download(job_preset, remote_project)

        # submit the job to the remote project
        self._log.info(u"Submitting job " + job_preset.get_name() + u" to existing remote project " +
                       remote_project.get_name())

        client.submit_project(gm_project, skip_upload=True, skip_auto_download=skip_auto_download)

        self._log.info(u"Job submitted - See Render Manager for details")

    def update_remote_project_status(self, remote_project):

        self._log.info(u"Fetching project status for project \"" + remote_project.get_name() + u"\"...")

        http_client = self._envoy_client.http_client
        url = self._envoy_client.url + u"/project-status/" + remote_project.get_name()

        try:
            resp = http_client.request('get', url)
        except Exception as e:
            msg = u"Could not fetch project status for project \"" + remote_project.get_name() + u"\". " + \
                  get_exception_with_traceback(e)
            self._log.error(msg)
            raise APIClientError(msg)
        else:
            if resp.status_code == 200:
                self._log.info(u"Fetched project status for project \"" + remote_project.get_name() + u"\".")

                json = resp.json()

                details = json.get(u'Details')

                total_bytes = 0
                total_bytes_done = 0
                files = set()

                if len(details.items()) > 0:
                    for detail_key, detail in details.items():
                        total_bytes_for_detail = detail.get(u'BytesTotal', 1)
                        bytes_done_for_detail = detail.get(u'BytesDone', 0)

                        total_bytes += total_bytes_for_detail
                        total_bytes_done += bytes_done_for_detail
                        files.add(pathlib.Path(detail_key))
                else:
                    # It is possible that the project has already been uploaded and Envoy is not bothering to show the
                    # project status for it any more.  Check to see if there are any files uploaded for this project.
                    uploaded_files = self.get_remote_project_files(remote_project.get_name(), ignore_cache=True)
                    if len(uploaded_files) >= 0:
                        for file in uploaded_files:
                            total_bytes += file.get_size()
                            total_bytes_done += file.get_size()
                            files.add(pathlib.Path(file.get_key()))

                remote_project.set_size(total_bytes)
                remote_project.set_total_bytes_done_uploading(total_bytes_done)
                remote_project.set_file_list(files)

            if resp.status_code == 404:
                msg = u"404: {0} not found".format(url)
                self._log.error(msg)
                raise APIClientError(msg)

    def get_cached_root_directories(self):
        return list(map(lambda x: x.get_name(), self._remote_projects_cache.get_value()))

    def get_root_directories(self, ignore_cache = False):
        self._log.info(u"Getting root directories. ignore_cache=" + unicode(ignore_cache))

        if ignore_cache or not self._remote_projects_cache.is_cached():
            self.get_remote_projects(ignore_cache=True)

        return self.get_cached_root_directories()

    def get_remote_projects(self, ignore_cache=False):
        self._log.info(u"Getting Envoy Projects. ignore_cache=" + unicode(ignore_cache))

        if ignore_cache or not self._remote_projects_cache.is_cached():
            url = self._envoy_client.url + u'/projects'
            self._log.info(u"Fetching existing remote project directories from \"" + url + u"\"")

            try:
                import requests
                r = requests.get(url)
                json = r.json()
                projects = json.get(u"projects", [])

                envoy_projects = []

                for project in projects:
                    envoy_projects.append(EnvoyProject(project.get(u'name', u''),
                                                       project.get(u'deleting', False),
                                                       project.get(u'num_files', 0),
                                                       project.get(u'total_size', 0),
                                                       project.get(u'auto_downloading', False),
                                                       project.get(u'watched_list', [])))

                projects = sorted(envoy_projects, key=lambda s: s.get_name().lower())

            except Exception as e:
                self._log.error(u"Could not make request to envoy API. " + unicode(e))
                self.sign_out()
                projects = []

            self._remote_projects_cache.set_value(projects)

        remote_projects = self._remote_projects_cache.get_value()
        self._log.info(u"Returning " + unicode(len(remote_projects))+ u" Envoy Project(s).")
        return remote_projects

    def get_product_versions(self, product, ignore_cache = False):
        if ignore_cache or not self._product_resolver.is_cached():
            # get product resolver
            resolver = self._envoy_client.get_product_resolver()

            # cache all products
            self._product_resolver.set_value(resolver.get_all_types())

        # sorted by latest version to oldest version
        return sorted(self._product_resolver.get_value().get(product, {}).get(u"versions", []),
                      key=lambda s: s.lower(), reverse=True)

    def get_closest_matching_product_version(self, product, product_version, ignore_cache = False):
        product_versions = self.get_product_versions(product, ignore_cache=ignore_cache)

        if not len(product_versions):
            return None

        # first check to see if the provided version exactly matches a version option
        for version_option in product_versions:
            if version_option == product_version:
                return version_option

        # if not then we are down to heuristics
        import difflib
        return difflib.get_close_matches(product_version, product_versions)[0]

    def get_remote_project_files(self, project_name, ignore_cache=False):
        """ Returns a list of Envoy File objects, if project does not exist it returns an empty list. """

        if ignore_cache or project_name not in self._project_files_dictionary_cache:
            import requests

            try:
                r = requests.get(self._envoy_client.url + u'/project/' + project_name + u'/Files')
            except Exception as e:
                raise APIClientError(unicode(e))

            json = r.json()
            all_files = json.get(u'project_files', {}).get(u'all_files', [])

            envoy_files = []
            for file in all_files:
                envoy_files.append(EnvoyFile(
                    file.get(u'CheckSum', u''),
                    file.get(u'LastModified', u''),
                    file.get(u'Name', u''),
                    file.get(u'Key', u''),
                    file.get(u'Size', 0),
                    file.get(u'FmtName', u''),
                ))

            self._project_files_dictionary_cache[project_name] = envoy_files

        return self._project_files_dictionary_cache[project_name]

    def clear_project_files_cache(self):
        self._project_files_dictionary_cache.clear()

    def get_available_credits(self, ignore_cache=False):
        self._log.info(u"Getting available credits...")

        if ignore_cache or not self._cached_available_credits.is_cached():
            url = u'{0}/credits-info'.format(self._envoy_client.url)
            self._log.info(u"Fetching available credits from " + url)

            try:
                resp = self._envoy_client.http_client.request('get', url)
                self._log.info(u"Finished request for available credits")
            except Exception as e:
                msg = u"Could not fetch available credits list: " + get_exception_with_traceback(e)
                self._log.error(msg)
                raise APIClientError(msg)
            else:
                if resp.status_code == 200:
                    content = resp.json()

                    self._log.info(u"Caching fetched available credits value...")
                    available_credits = content.get(u'credits_available', self._cached_available_credits.get_value())

                    self._cached_available_credits.set_value(available_credits)

                if resp.status_code == 404:
                    msg = u"404: {0} not found".format(url)
                    self._log.error(msg)
                    raise APIClientError(msg)

        cached_available_credits = self._cached_available_credits.get_value()
        self._log.info(unicode(cached_available_credits) + u" available credits.")
        return cached_available_credits

    def get_products(self, ignore_cache=False):
        self._log.info(u"Getting products list...")

        if not self.is_user_signed_in():
            msg = u"Must be signed-in to get products list."
            self._log.error(msg)
            raise NotSignedInError(msg)

        if ignore_cache or not self._products_cache.is_cached():
            url = u'{0}/products'.format(self._envoy_client.url)

            self._log.info(u"Fetching products list from " + url)

            try:
                resp = self._envoy_client.http_client.request('get', url)
            except Exception as e:
                msg = u"Could not fetch products list: " + get_exception_with_traceback(e)
                self._log.error(msg)
                raise APIClientError(msg)
            else:
                if resp.status_code == 200:
                    content = resp.json()

                    def parse_product(json):
                        return Product(json.get(u'id', -1),
                                       json.get(u'name', u''),
                                       json.get(u'app_type', u''),
                                       json.get(u'version', u''))

                    self._log.info(u"Caching fetched products...")
                    self._products_cache.set_value(list(map(lambda x: parse_product(x), content)))

                if resp.status_code == 404:
                    msg = u"404: {0} not found".format(url)
                    self._log.error(msg)
                    raise APIClientError(msg)

        cached_products = self._products_cache.get_value()
        self._log.info(u"Returning " + unicode(len(cached_products)) + u" cached products...")
        return cached_products

    def get_products_with_app_type(self, app_type, ignore_cache=False):
        products = self.get_products(ignore_cache=ignore_cache)
        return list(filter(lambda x: x.get_app_type() == app_type, products))

    def get_product_app_types(self, ignore_cache=False):
        return sorted(list(set(map(lambda x: x.get_app_type(), self.get_products(ignore_cache=ignore_cache)))),
                      key=lambda s: s.lower(), reverse=True)

    def get_user_info(self, ignore_cache=False):
        self._log.info(u"Getting User info. ignore_cache=" + unicode(ignore_cache))

        if ignore_cache or not self._cached_user_info.is_cached():
            url = u'{0}/user-info'.format(self._envoy_client.url)
            self._log.info(u"Fetching user info from url \"" + url + u"\"")

            try:
                resp = self._envoy_client.http_client.request(u'get', url)
            except Exception as e:
                msg = u"Could not fetch user info: " + get_exception_with_traceback(e)
                self._log.error(msg)
                raise APIClientError(msg)
            else:
                if resp.status_code == 200:
                    content = resp.json().get(u'data')

                    max_through_put = content.get('max_throughput', 9999)
                    job_throughput_limit = content.get('job_throughput_limit', 60)

                    user_info = UserInfo(max_through_put, job_throughput_limit)

                    self._log.info(u"Caching user info...")
                    self._cached_user_info.set_value(user_info)

                if resp.status_code == 404:
                    msg = u"404: {0} not found".format(url)
                    self._log.error(msg)
                    raise APIClientError(msg)

        self._log.info(u"Returning User info.")
        return self._cached_user_info.get_value()

    def get_machines(self,
                     app,
                     operation,
                     plugins=None,
                     ignore_cache = False):

        cached_machine_options = self._cached_machine_options.get_value()

        if plugins is None:
            plugins = []

        _app_cahce_id = app + ''.join(plugins)

        if _app_cahce_id in cached_machine_options:
            app_machine_options = cached_machine_options[_app_cahce_id]
            if operation in app_machine_options:
                operation_machine_options = app_machine_options[operation]

                if not ignore_cache:
                    return operation_machine_options

        try:
            self._log.info(u"Fetching machine types for product \"" + app + u"\" (operation=\"" + operation +
                           u"\" plugins=[" + ','.join(plugins) + u")")

            cpu_machine_options = self._envoy_client.get_machines(app, operation, False, *plugins)
            gpu_machine_options = self._envoy_client.get_machines(app, operation, True, *plugins)

        except APIError as e:
            msg = unicode(e.user_message)
            self._log.error(msg)
            raise APIClientError(message=msg)

        def parse_machine_option(machine_option, is_gpu):
            return MachineOption(machine_option[u'id'], machine_option[u'name'], machine_option[u'is_default'], is_gpu)

        cpu_machine_options_list = list(map(lambda x: parse_machine_option(x, False), cpu_machine_options))
        gpu_machine_options_list = list(map(lambda x: parse_machine_option(x, True), gpu_machine_options))
        combined_machine_options = cpu_machine_options_list + gpu_machine_options_list

        # It is possible no machine options were returned, in that case we can still let users choose between their
        # default machine options (cpu vs gpu)
        if len(combined_machine_options) <= 0:
            combined_machine_options.append(MachineOption(u"default_machine_cpu",
                                                          u"Default CPU Machine (If Supported)",
                                                          u"Use your default cpu machine (As defined on your profile page in the portal)",
                                                          False))
            combined_machine_options.append(MachineOption(u"default_machine_gpu",
                                                          u"Default GPU Machine (If Supported)",
                                                          u"Use your default gpu machine (As defined on your profile page in the portal)",
                                                          True))

        app_machine_options = cached_machine_options.get(_app_cahce_id, {})
        app_machine_options[operation] = combined_machine_options
        cached_machine_options[_app_cahce_id] = app_machine_options

        self._cached_machine_options.set_value(cached_machine_options)

        return combined_machine_options

    def _is_gpu_enabled_machine(self, machine_id):

        # get all cached machine types
        cached_machine_options_dict = self._cached_machine_options.get_value()

        cached_machine_options = set()
        for app, operations in cached_machine_options_dict.iteritems():
            for operation, machine_options in operations.iteritems():
                cached_machine_options.update(machine_options)

        for machine_option in cached_machine_options:
            if machine_option.get_id() == machine_id:
                return machine_option.is_gpu_machine()

        return ValueError(u"Could not find cached machine option with id '" + unicode(machine_id) + u"'")
