__all__ = 'ProjectAttribute'

from abc import abstractmethod

from .errors import RejectedTransitionInputError
from gm_meta_plugin.plugin_accessor import PluginAccessor


class ProjectAttribute(PluginAccessor):

    def __init__(self,
                 plugin,
                 project_attribute_id,
                 attribute,
                 transitions,
                 compatible_job_definitions):
        PluginAccessor.__init__(self, plugin)
        self._id = project_attribute_id
        self._attribute = attribute
        self._transitions = transitions
        self._compatible_job_definitions = compatible_job_definitions

    def get_id(self):
        return self._id

    def get_attribute(self):
        return self._attribute

    @abstractmethod
    def get_value(self):
        raise NotImplementedError

    @abstractmethod
    def set_value(self, value):
        raise NotImplementedError

    def has_transitions(self):
        return len(self._transitions) > 0

    def get_transitions(self):
        return self._transitions

    def has_compatible_job_definitions(self):
        return len(self._compatible_job_definitions) > 0

    def get_compatible_job_definitions(self):
        return self._compatible_job_definitions

    def transition(self, input, force_transition=False):
        for transition in self._transitions:
            try:
                project_attribute = transition.transition(input)
                return project_attribute
            except RejectedTransitionInputError:
                pass

        # if there is only one possible transition then we can either
        if len(self.get_transitions()) == 1:
            transition = self.get_transitions()[0]
            # force the transition
            if force_transition:
                return transition.get_project_attribute()

            # or return a more specific error message if one exists
            message = transition.get_rejection_message()
        else:
            message = "Project attribute '" + self.get_attribute().get_display_name() + " (" + self.get_id() + ")" + \
                      "' does not have a transition that accepts the input value '" + unicode(input) + "'"

        raise RejectedTransitionInputError(message)

    def get_children(self):
        return list(map(lambda x: x.get_project_attribute(), self.get_transitions()))

    def get_descendants(self):
        """ Returns all children and all the children's children recursively """
        children = self.get_children()
        return children + list(map(lambda x: x.get_descendants(), children))
