__all__ = 'MainWindow'

from PySide2 import QtCore
from PySide2 import QtWidgets

from gm_qt_plugin import qt_constants, qt_utils
from gm_qt_plugin.widgets import MainUI, SignInWidget
from gm_qt_plugin.qt_plugin_accessor import QtPluginAccessor
from gm_qt_plugin.qt_user_interface import QtUserInterface


class MainWindow(QtWidgets.QDialog, QtPluginAccessor):

    def __init__(self, plugin, reset_callback, parent=None):
        QtWidgets.QDialog.__init__(self, parent)
        QtPluginAccessor.__init__(self, plugin)

        self._reset_callback = reset_callback

        # add the MainWindow to the plugins user interface class
        plugin.get_user_interface().set_main_window(self)

        # load custom fonts
        QtUserInterface.load_fonts()

        # Remove the window help button that only appears on Windows OS
        self.setWindowFlags(self.windowFlags() ^ QtCore.Qt.WindowContextHelpButtonHint)

        # Set the window title to the plugin name
        self.setWindowTitle(plugin.get_name())

        self.setWindowIcon(plugin.get_user_interface().get_icon())

        self._closed = False

        # The main window has two main states:
        # 1. The user has yet to sign in
        #    - Display a simple sign-in layout
        #
        # 2. The user is signed in
        #    - Display the main plugin user interface
        #
        # The layout for (2) can not be generated until the user is signed in so the creation of that layout is deferred
        # until after the user has signed-in.

        # the container widget is the widget sub widgets (1) & (2) will be parented to. This is done so that
        self._container_widget = QtWidgets.QStackedWidget()
        self._container_widget.setProperty(qt_constants.STYLE_PROPS.MAIN_CONTAINER_WIDGET, True)

        # Override the style object and apply it to the container, this way any styles applied by the parent do not
        # affect our widgets.
        self._style_override = QtWidgets.QStyleFactory().create('windows')
        self._container_widget.setStyle(self._style_override)

        self._sign_in_widget = None
        self._main_ui_widget = None

        self._sign_in_widget_index = None
        self._main_ui_widget_index = None

        self._layout = QtWidgets.QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.addWidget(self._container_widget)

        self.display_signed_out_ui()

    def is_closed(self):
        return self._closed

    def get_sign_in_widget(self):

        if not self._sign_in_widget:
            self._sign_in_widget = SignInWidget(self.get_plugin())

        return self._sign_in_widget

    def get_main_ui_widget(self):

        if not self._main_ui_widget:
            self._main_ui_widget = MainUI(self.get_plugin())

        return self._main_ui_widget

    def get_sign_in_widget_index(self):

        if self._sign_in_widget_index is None:
            self._sign_in_widget_index = self._container_widget.addWidget(self.get_sign_in_widget())

        return self._sign_in_widget_index


    def get_main_ui_widget_index(self):

        if self._main_ui_widget_index is None:
            self._main_ui_widget_index = self._container_widget.addWidget(self.get_main_ui_widget())

        return self._main_ui_widget_index

    def display_signed_in_ui(self):
        # set min and max values (fixed) for both width and height
        self.setMinimumWidth(MainUI.MIN_WINDOW_WIDTH)
        self.setMinimumHeight(MainUI.MIN_WINDOW_HEIGHT)

        self.setMaximumWidth(MainUI.MAX_WINDOW_WIDTH)
        self.setMaximumHeight(MainUI.MAX_WINDOW_HEIGHT)

        # perform the resize
        self.resize(MainUI.INITIAL_WINDOW_WIDTH, MainUI.INITIAL_WINDOW_HEIGHT)

        qt_utils.center_window(self)

        # load style sheet
        # self.get_plugin().get_user_interface().set_style_sheet(main_window)
        self.setWindowFlags(self.windowFlags() & ~QtCore.Qt.FramelessWindowHint)
        self.setWindowFlags(self.windowFlags() | QtCore.Qt.WindowMaximizeButtonHint)
        self.setWindowFlags(self.windowFlags() | QtCore.Qt.WindowMinimizeButtonHint)
        self.show()

        self.get_plugin().get_user_interface().set_style_sheet(self._container_widget)
        self._container_widget.setCurrentIndex(self.get_main_ui_widget_index())

    def display_signed_out_ui(self):
        # set min and max values (fixed) for both width and height
        self.setFixedWidth(SignInWidget.WINDOW_WIDTH)
        self.setFixedHeight(SignInWidget.WINDOW_HEIGHT)

        # perform the resize
        self.resize(SignInWidget.WINDOW_WIDTH, SignInWidget.WINDOW_HEIGHT)

        qt_utils.center_window(self)

        # load style sheet
        # self.get_plugin().get_user_interface().set_style_sheet(main_window, qt_constants.WhiteTheme)
        self.setWindowFlags(self.windowFlags() | QtCore.Qt.FramelessWindowHint)

        self.get_plugin().get_user_interface().set_style_sheet(self._container_widget, qt_constants.WhiteTheme)
        self._container_widget.setCurrentIndex(self.get_sign_in_widget_index())

    def reset(self):
        self._reset_callback()

    def close_window(self):
        self._closed = True
        self.close()
        QtUserInterface.unload_fonts()
