from PySide2 import QtCore
from PySide2 import QtGui
from PySide2 import QtWidgets

import pathlib

import math
import re
import sys

import gm_meta_plugin
from gm_meta_plugin.gridmarkets import constants as api_constants
from gm_meta_plugin.attribute import *
from gm_meta_plugin.attribute_types import *

from . import qt_constants


def save_logs_to_file(plugin):
    file_dialog = QtWidgets.QFileDialog()
    file_dialog.setFileMode(QtWidgets.QFileDialog.AnyFile)
    file_dialog.setViewMode(QtWidgets.QFileDialog.Detail)
    file_dialog.setAcceptMode(QtWidgets.QFileDialog.AcceptSave)
    file_dialog.setNameFilter("*.txt")
    file_dialog.setDefaultSuffix('.txt')
    file_dialog.setLabelText(QtWidgets.QFileDialog.Accept, "Save")
    file_dialog.setLabelText(QtWidgets.QFileDialog.Reject, "Cancel")
    file_dialog.setLabelText(QtWidgets.QFileDialog.FileName, "File Name")
    file_dialog.setDirectory(gm_meta_plugin.utils.get_desktop_path())

    if file_dialog.exec_():
        file_names = file_dialog.selectedFiles()

        # only one file can be selected anyway
        save_file = file_names[0]

        logging_coordinator = plugin.get_logging_coordinator()
        logging_coordinator.save_logs_to_file(save_file)


def get_product_icon(product):
    if product == api_constants.PRODUCTS.BLENDER:
        return qt_constants.BLENDER_ICON
    elif product == api_constants.PRODUCTS.MAYA:
        return qt_constants.MAYA_ICON
    elif product == api_constants.PRODUCTS.MOE:
        return qt_constants.MOE_ICON
    elif product == api_constants.PRODUCTS.VRAY:
        return qt_constants.VRAY_ICON
    else:
        return qt_constants.EMPTY_ICON


def mix_colors(color1, color2, factor):
    """
    :type color1: QColor
    :type color2: QColor
    :rtype: QColor
    """

    def blend_color_value(v1, v2, r):
        return math.sqrt((1 - r) * math.pow(v1, 2) + r * math.pow(v2, 2))

    def blend_alpha_value(v1, v2, r):
        return (1 - r) * v1 + r * v2

    combined_color = QtGui.QColor()
    combined_color.setRed(blend_color_value(color1.red(), color2.red(), factor))
    combined_color.setGreen(blend_color_value(color1.green(), color2.green(), factor))
    combined_color.setBlue(blend_color_value(color1.blue(), color2.blue(), factor))
    combined_color.setAlpha(blend_alpha_value(color1.blue(), color2.blue(), factor))

    return combined_color


def add_widgets_for_attribute(plugin, attribute, widgets, widget_id):
    from gm_qt_plugin.widgets import input_widgets

    display_name = attribute.get_display_name()
    description = attribute.get_description()
    attribute_type = attribute.get_type()
    default = attribute.get_default_value()
    subtype_kwargs = attribute.get_subtype_kwargs()

    if attribute_type == AttributeType.NULL:
        return
    elif attribute_type == AttributeType.STRING:
        string_attribute = attribute
        subtype = string_attribute.get_subtype()
        placeholder = string_attribute.get_placeholder()

        max_length = string_attribute.get_max_length()

        if subtype == StringSubtype.NONE.value:
            line_edit = input_widgets.LineEdit()

            if max_length:
                line_edit.setMaxLength(max_length)

            if placeholder:
                line_edit.setPlaceholderText(placeholder)

            line_edit.setText(default)

            if description:
                line_edit.setToolTip(description)

            # Set max length
            if string_attribute.get_max_length() is not None:
                line_edit.setMaxLength(string_attribute.get_max_length())

            widgets[widget_id] = line_edit

            # if the attribute key indicates it is a Project Name attribute then we need also need to add a dynamic
            # QComboWidget which lets the user choose from an existing project name
            # todo
            """
            if attribute.get_key() == api_constants.API_KEYS.PROJECT_NAME:

                def get_remote_root_directories(self, context):
                    plugin = PluginFetcher.get_plugin_if_initialised()

                    if plugin is None:
                        return []

                    api_client = plugin.get_api_client()
                    projects = api_client.get_cached_root_directories()

                    return list(map(lambda project: (project, project, ''), projects))

                project_name_combo_box = QtWidgets.QComboBox()
                widgets[widget_id + constants.REMOTE_SOURCE_SUFFIX] = QtWidgets.QComboBox()
            """

        elif subtype == StringSubtype.FRAME_RANGES.value:

            # treat as simple line edit for now
            frame_range_input = input_widgets.FrameRangeInput()

            frame_range_input.set_value(default)
            widgets[widget_id] = frame_range_input

        elif subtype == StringSubtype.PATH.value:

            # get the file mode
            file_mode = subtype_kwargs[api_constants.SUBTYPE_KEYS.STRING.PATH.FILE_MODE]
            if file_mode == api_constants.SUBTYPE_KEYS.STRING.PATH.FILE_PATH:
                file_mode = QtWidgets.QFileDialog.AnyFile
            elif file_mode == api_constants.SUBTYPE_KEYS.STRING.PATH.DIR_PATH:
                file_mode = QtWidgets.QFileDialog.Directory
            else:
                raise ValueError("Unrecognised file mode for PathInput")

            file_input = input_widgets.PathInput(file_mode=file_mode)
            file_input.set_value(default)

            if description:
                file_input.set_tool_tip(description)

            if placeholder:
                file_input.set_placeholder_text(placeholder)

            widgets[widget_id] = file_input

            # use simple line edit for remote file input
            if file_mode == api_constants.SUBTYPE_KEYS.STRING.PATH.FILE_PATH:
                line_edit = input_widgets.LineEdit()
                line_edit.setPlaceholderText(placeholder)

                if max_length:
                    line_edit.setMaxLength(max_length)

                line_edit.setText(default)
                widgets[widget_id + qt_constants.REMOTE_SOURCE_SUFFIX] = line_edit

        else:
            raise ValueError("Unrecognised String subtype '" + subtype + "'.")

    elif attribute_type == AttributeType.ENUM:
        enum_attribute = attribute
        subtype = enum_attribute.get_subtype()
        combo_box = input_widgets.ComboBox()

        if subtype == EnumSubtype.PRODUCT_VERSIONS.value:

            product = subtype_kwargs.get(api_constants.SUBTYPE_KEYS.ENUM.PRODUCT_VERSIONS.PRODUCT)

            match = subtype_kwargs.get(api_constants.SUBTYPE_KEYS.ENUM.PRODUCT_VERSIONS.MATCH)
            match = p = re.compile(match) if match is not None else None

            versions = plugin.get_api_client().get_product_versions(product)

            if match is not None:
                versions = [s for s in versions if p.match(s)]

            for version in versions:
                combo_box.addItem(version, version)

            # if the product is in the supported applications list
            if product in plugin.get_supported_applications():
                # then we can get current version of the application and use it as the default item
                application_attribute_source = plugin.get_application_attribute_source()
                items, default_item_index = application_attribute_source.get_enum_items(None, None, attribute.get_key())
                default_version = items[default_item_index]

                # attempt to set the default item to the version returned by the app attribute source.
                # If the default version does not match any version in the versions list then this does nothing.
                combo_box.setCurrentText(default_version.get_key())

        elif subtype == EnumSubtype.MACHINE_TYPE.value:

            app = subtype_kwargs.get(api_constants.SUBTYPE_KEYS.ENUM.MACHINE_TYPE.PRODUCT)
            operation = subtype_kwargs.get(api_constants.SUBTYPE_KEYS.ENUM.MACHINE_TYPE.OPERATION)
            use_gpu_default = subtype_kwargs.get(api_constants.SUBTYPE_KEYS.ENUM.MACHINE_TYPE.USE_GPU_DEFAULT) == "True"
            machine_options = plugin.get_api_client().get_machines(app, operation)

            for i, machine_option in enumerate(machine_options):

                combo_box.addItem(machine_option.get_name(), machine_option.get_id())

                # give the default machine id 0 so it starts selected
                if machine_option.is_default() and machine_option.is_gpu_machine() == use_gpu_default:
                    combo_box.setCurrentIndex(i)

        elif subtype == EnumSubtype.APP_INFERRED.value:
            application_attribute_source = plugin.get_application_attribute_source()

            def rebuild_items(index=None):
                old_items = [combo_box.itemText(i) for i in range(combo_box.count())]
                old_index = combo_box.currentIndex()

                combo_box.clear()

                items, default_item_index = application_attribute_source.get_enum_items(None, None, attribute.get_key())
                for item in items:
                    combo_box.addItem(item.get_display_name(), item.get_key())

                if old_items == map(lambda x: x.get_display_name(), items):
                    combo_box.setCurrentIndex(old_index)
                else:
                    combo_box.setCurrentIndex(default_item_index)

                    # flash the combo_box to indicate to the user that the items have changed
                    user_interface = plugin.get_user_interface()
                    colors = user_interface.get_theme()
                    original_color = QtGui.QColor(colors.FOREGROUND_COLOR)
                    flash_color = QtGui.QColor(colors.BRAND_COLOR)
                    flash_widget_color(combo_box, original_color, flash_color)

            rebuild_items(None)
            combo_box.set_update_items_callback(rebuild_items)

        else:
            enum_items = enum_attribute.get_items()
            for enum_item in enum_items:
                combo_box.addItem(enum_item.get_display_name(), enum_item.get_key())
            combo_box.setCurrentText(default)

        widgets[widget_id] = combo_box

    elif attribute_type == AttributeType.BOOLEAN:
        check_box = input_widgets.CheckBox()
        check_box.setChecked(default)

        if description:
            check_box.setToolTip(description)

        widgets[widget_id] = check_box

    elif attribute_type == AttributeType.INTEGER:
        integer_attribute = attribute

        if integer_attribute.get_subtype() == IntegerSubtype.INSTANCES.value:
            user_info = plugin.get_api_client().get_user_info()

            integer_attribute._default_value = user_info.get_job_throughout_limit()

            machine_instances_input = input_widgets.MachineInstancesInput()
            machine_instances_input.set_value(user_info.get_job_throughout_limit())
            machine_instances_input.set_maximum(user_info.get_max_throughout())
            machine_instances_input.set_minimum(1)

            widgets[widget_id] = machine_instances_input
        else:
            spin_box = input_widgets.SpinBox()

            if description:
                spin_box.setToolTip(description)

            spin_box.setMaximum(qt_constants.MAX_INT)  # default spinBox max is 99
            spin_box.setValue(default)
            widgets[widget_id] = spin_box


def delete_layout_items(layout):
    while layout.count():
        layout_item = layout.takeAt(0)

        if not layout_item:
            break

        # if the item is a sub layout recurse and delete the sub layout first
        if isinstance(layout_item.layout(), QtWidgets.QLayout):
            sub_layout = layout_item.layout()
            delete_layout(sub_layout)

        # if the item is a widget then delete the widget
        elif isinstance(layout_item.widget(), QtWidgets.QWidget):
            widget = layout_item.widget()
            layout.removeWidget(widget)
            widget.hide()
            widget.deleteLater()

        # if the item is a widget then delete the widget
        elif isinstance(layout_item.spacerItem(), QtWidgets.QSpacerItem):
            spacer = layout_item.spacerItem()
            del spacer

        del layout_item


def delete_layout(layout):

    # re-parent the layout to a temporary widget. This way it will be removed from it's previous parent
    temp_widget = QtWidgets.QWidget()
    temp_widget.setLayout(layout)

    # remove all the sub items in the layout
    delete_layout_items(layout)

    # mark everything for deletion
    temp_widget.deleteLater()
    layout.deleteLater()
    del temp_widget
    del layout


def open_url(url):
    import webbrowser
    webbrowser.open(url)


def get_project_files(plugin, widgets, files=None, project_attribute = None):
    root = plugin.get_api_client().get_cached_api_schema().get_project_definition_schema().get_root_project_attribute()

    # If no files array is provided create one
    files = [] if files is None else files

    # if project attribute is none use the root attribute
    project_attribute = root if project_attribute is None else project_attribute

    attribute = project_attribute.get_attribute()

    if attribute.get_type() == AttributeType.NULL:
        return files

    value = project_attribute.get_value(widgets=widgets)

    # if the attribute is a file path
    if attribute.get_type() == AttributeType.STRING and \
            attribute.get_subtype() == StringSubtype.PATH.value and \
            attribute.get_subtype_kwargs().get(api_constants.SUBTYPE_KEYS.STRING.PATH.FILE_MODE) == api_constants.SUBTYPE_KEYS.STRING.PATH.FILE_PATH:
        # add it's value to the files array
        files.append(pathlib.Path(value))

    return get_project_files(plugin, widgets, files, project_attribute.transition(value))


def remove_widgets_from_layout(layout):
    while True:
        child = layout.takeAt(0)

        if child is 0:
            return

        # The documentation says it should never be none but sometimes it is...
        if child is None:
            return

        elif child.widget() is not None:
            widget = child.widget()
            layout.removeWidget(widget)
            widget.hide()
        elif child.layout() is not None:
            remove_widgets_from_layout(child.layout())
            layout.removeItem(child)


def show_all_widgets_in_layout(layout):
    i = 0
    while i < layout.count():
        child = layout.itemAt(i)

        if child is 0:
            pass
        elif child is None:
            pass
        elif child.widget() is not None:
            widget = child.widget()
            widget.show()
        elif child.layout() is not None:
            show_all_widgets_in_layout(child.layout())

        i = i + 1


def transfer_all_widgets_in_layout(layout, new_layout):
    i = 0
    while i < layout.count():
        child = layout.itemAt(i)

        if child is 0:
            pass
        elif child is None:
            pass
        elif child.widget() is not None:
            widget = child.widget()
            widget.hide()
            new_layout.addWidget(widget)
        elif child.layout() is not None:
            show_all_widgets_in_layout(child.layout())

        i = i + 1


def center_window(window):
    frameGm = window.frameGeometry()

    # get the active screen
    screen = QtWidgets.QApplication.desktop().screenNumber(QtWidgets.QApplication.desktop().cursor().pos())

    # get the center of the active screen
    centerPoint = QtWidgets.QApplication.desktop().screenGeometry(screen).center()

    # move window to the centre of the active screen
    frameGm.moveCenter(centerPoint)
    window.move(frameGm.topLeft())


def flash_widget_color(widget, original_color, flash_color):

    # if a flash animation is already in progress do not continue (may cause crash)
    if hasattr(widget, '___flash_effect') and widget.___flash_effect is not None:
        return

    widget.___flash_effect = QtWidgets.QGraphicsColorizeEffect(widget)
    widget.___flash_effect.setColor(flash_color)
    widget.setGraphicsEffect(widget.___flash_effect)
    widget.___a = QtCore.QPropertyAnimation(widget.___flash_effect, "strength")
    widget.___a.setDuration(3000)
    widget.___a.setStartValue(0)
    widget.___a.setKeyValueAt(0.2, 1)
    widget.___a.setKeyValueAt(0.7, 1)
    widget.___a.setEndValue(0)
    widget.___a.setEasingCurve(QtCore.QEasingCurve.OutCubic)
    widget.___a.start(QtCore.QPropertyAnimation.DeleteWhenStopped)

    def clean_up():
        widget.___flash_effect.deleteLater()
        widget.___flash_effect = None

    widget.___a.finished.connect(clean_up)
