__all__ = 'JobPresetAttribute'

from PySide2 import QtCore
from PySide2 import QtGui
from PySide2 import QtWidgets

import gm_meta_plugin
from gm_qt_plugin.widgets.input_widgets import Constant, CheckBox
from gm_qt_plugin.qt_plugin_accessor import QtPluginAccessor
from gm_qt_plugin import qt_utils, qt_constants
from gm_qt_plugin.widgets.elided_label import ElidedLabel
from gm_qt_plugin.widgets.input_widgets.combo_box import ComboBox


class JobPresetAttribute(QtWidgets.QWidget, gm_meta_plugin.JobPresetAttribute, QtPluginAccessor):

    def __init__(self, plugin, parent_preset, job_attribute, inference_source=None, value=None, parent=None):
        QtWidgets.QWidget.__init__(self, parent=parent)
        gm_meta_plugin.JobPresetAttribute.__init__(self,
                                                   parent_preset,
                                                   job_attribute,
                                                   inference_source=inference_source,
                                                   value=value)
        QtPluginAccessor.__init__(self, plugin)

        job_attribute = self.get_job_attribute()
        attribute = job_attribute.get_attribute()
        prop_id = job_attribute.get_attribute().get_key()
        inferences_sources = self.get_job_attribute().get_inference_sources()

        self._initialised = False

        ''' Create Widgets '''

        # display name label
        self._display_name_label = ElidedLabel(attribute.get_display_name())
        self._display_name_label.setMinimumWidth(130)
        self._display_name_label.setProperty('h5', True)

        # attribute key label
        self._attribute_key_label = ElidedLabel(self.get_key())
        self._attribute_key_label.setMinimumWidth(125)

        # various input widgets
        self._widgets = {}
        qt_utils.add_widgets_for_attribute(self.get_plugin(),
                                           job_attribute.get_attribute(),
                                           self._widgets,
                                           prop_id)

        # create the inference source widget
        self._inference_source_widget = CheckBox()

        # create a stacked widget to hold the difference input widgets
        self._stacked_widget = QtWidgets.QStackedWidget()

        # validate the inference source options ( this are temporary checks for v1)
        is_application_inferred = gm_meta_plugin.InferenceSource.get_application_inference_source() in inferences_sources
        is_user_inferred = gm_meta_plugin.InferenceSource.get_user_defined_inference_source() in inferences_sources
        is_constant_inferred = gm_meta_plugin.InferenceSource.get_constant_inference_source() in inferences_sources
        is_project_inferred = gm_meta_plugin.InferenceSource.get_project_inference_source() in inferences_sources

        if is_constant_inferred | is_project_inferred:
            if is_application_inferred | is_user_inferred:
                raise ValueError()

        if inferences_sources[0] == gm_meta_plugin.InferenceSource.get_user_defined_inference_source():
            self._inference_source_widget.setCheckState(QtCore.Qt.Checked)

        # if the job preset attribute is not either app inferred or user inferred  then don't show the input
        if not is_application_inferred or not is_user_inferred:
            self._inference_source_widget.setProperty('hidden', True)
            self._inference_source_widget.setEnabled(False)

        self._inference_source_widget.input_value_changed.connect(self.inference_source_option_changed)

        if gm_meta_plugin.InferenceSource.get_user_defined_inference_source() in inferences_sources:
            self._user_input_widget = self._widgets[prop_id]
            self._stacked_widget.addWidget(self._user_input_widget)

        if gm_meta_plugin.InferenceSource.get_constant_inference_source() in inferences_sources:
            self._constant_widget = Constant(job_attribute.get_attribute().get_default_value())
            self._constant_widget.setToolTip(attribute.get_description())
            self._stacked_widget.addWidget(self._constant_widget)

        if gm_meta_plugin.InferenceSource.get_application_inference_source() in inferences_sources:
            # deferred until requested as it requires eval which cannot be used in the constructor
            self._app_inferred_value = None

        if len(inferences_sources) == 1:
            self._inference_source_widget.setEnabled(False)

        """ Create Layouts """
        self._layout = QtWidgets.QHBoxLayout(self)
        self._layout.setContentsMargins(0,0,0,0)
        self._layout.addWidget(self._stacked_widget)

    def init(self):
        """ Call before adding to a layout """
        if not self._initialised:

            # add the app input widget to that stacked widget
            inferences_sources = self.get_job_attribute().get_inference_sources()
            if gm_meta_plugin.InferenceSource.get_application_inference_source() in inferences_sources:
                self._stacked_widget.addWidget(self.get_app_inferred_widget())

            # set the starting widget
            self._stacked_widget.setCurrentWidget(self.get_current_widget())

            self._initialised = True

    def update_app_inferred_values(self):
        current_widget = self._stacked_widget.currentWidget()

        # required to update combo box items
        if isinstance(current_widget, ComboBox):
            current_widget.update_items()

        if self.get_inference_source() == gm_meta_plugin.InferenceSource.get_application_inference_source():
            old_value = current_widget.get_value()
            new_value = unicode(self.eval(plugin=self.get_plugin()))
            current_widget.set_value(new_value)

            if old_value != new_value:
                user_interface = self.get_plugin().get_user_interface()
                theme = user_interface.get_theme()
                original_color = QtGui.QColor(theme.FOREGROUND_COLOR)
                flash_color = QtGui.QColor(theme.BRAND_COLOR)
                qt_utils.flash_widget_color(current_widget, original_color, flash_color)

    @QtCore.Slot(int)
    def inference_source_option_changed(self):
        inference_source = self.get_inference_source()

        if self._inference_source_widget.isChecked():
            self.set_inference_source(gm_meta_plugin.InferenceSource.get_user_defined_inference_source())
            self._stacked_widget.setCurrentWidget(self.get_user_input_widget())
        else:
            self.set_inference_source(gm_meta_plugin.InferenceSource.get_application_inference_source())
            self._stacked_widget.setCurrentWidget(self.get_app_inferred_widget())

        self.update_app_inferred_values()

    def get_display_name_label(self):
        return self._display_name_label

    def get_key_label(self):
        return self._attribute_key_label

    def get_constant_widget(self):
        return self._constant_widget

    def get_user_input_widget(self):
        return self._user_input_widget

    def get_app_inferred_widget(self):

        # we defer the creation of this widget as to set its value we need to use 'eval' which cannot be done in the
        # constructor
        if self._app_inferred_value is None:
            attribute = self.get_job_attribute().get_attribute()
            self._app_inferred_value = Constant(u"")
            value = unicode(self.eval(plugin=self.get_plugin()))
            self._app_inferred_value.set_value(value)
            self._app_inferred_value.setToolTip(attribute.get_description())

            def clicked():
                inferences_sources = self.get_job_attribute().get_inference_sources()
                is_user_inferred = gm_meta_plugin.InferenceSource.get_user_defined_inference_source() in inferences_sources

                if is_user_inferred:
                    # get the current app inferred value
                    app_inferred_value = unicode(self.eval(plugin=self.get_plugin()))
                    
                    self.get_user_input_widget().set_value(app_inferred_value)
                    self._inference_source_widget.set_value(True)

            self._app_inferred_value.clicked.connect(clicked)

        return self._app_inferred_value

    def get_inference_source_widget(self):
        return self._inference_source_widget

    def get_current_widget(self, project_inferred_value="Project inferred value"):
        inference_source = self.get_inference_source()

        if inference_source == gm_meta_plugin.InferenceSource.get_constant_inference_source():
            widget = self.get_constant_widget()

        elif inference_source == gm_meta_plugin.InferenceSource.get_application_inference_source():
            widget = self.get_app_inferred_widget()
                
        elif inference_source == gm_meta_plugin.InferenceSource.get_user_defined_inference_source():
            widget = self.get_user_input_widget()

        elif inference_source == gm_meta_plugin.InferenceSource.get_project_inference_source():
            widget = Constant(project_inferred_value)
            widget.setToolTip(self.get_job_attribute().get_attribute().get_description())
        else:
            raise ValueError("Unsupported inference source type: " + inference_source.get_display_name())

        return widget

    def get_value(self):
        inference_source = self.get_inference_source()

        if inference_source == gm_meta_plugin.InferenceSource.get_constant_inference_source():
            value = self.get_job_attribute().get_attribute().get_default_value()

        elif inference_source == gm_meta_plugin.InferenceSource.get_application_inference_source():
            value = self.get_app_inferred_widget().get_value()

        elif inference_source == gm_meta_plugin.InferenceSource.get_user_defined_inference_source():
            value = self.get_user_input_widget().get_value()

        elif inference_source == gm_meta_plugin.InferenceSource.get_project_inference_source():
            raise NotImplementedError
        else:
            raise ValueError("Unsupported inference source type: " + inference_source.get_display_name())

        return value

    def set_value(self, value):
        self.get_user_input_widget().set_value(value)

    def get_inference_source(self):
        """
        index = self.get_inference_source_widget().currentIndex()
        inference_source = self.get_job_attribute().get_inference_sources()[index]
        return inference_source
        """
        inference_source_widget = self.get_inference_source_widget()
        inferences_sources = self.get_job_attribute().get_inference_sources()

        if len(inferences_sources) == 1:
            return inferences_sources[0]
        else:
            if inference_source_widget.isChecked():
                return gm_meta_plugin.InferenceSource.get_user_defined_inference_source()
            else:
                return gm_meta_plugin.InferenceSource.get_application_inference_source()
