__all__ = 'SignInWidget'

from PySide2 import QtCore
from PySide2 import QtGui
from PySide2 import QtWidgets

import gm_meta_plugin

from gm_qt_plugin import qt_constants, qt_utils
from gm_qt_plugin.qt_none import QtNone
from gm_qt_plugin.worker_thread import WorkerThread
from gm_qt_plugin.qt_plugin_accessor import QtPluginAccessor
from gm_qt_plugin.widgets.push_button import PushButton
from gm_qt_plugin.widgets.input_widgets.line_edit import LineEdit

from gm_qt_plugin.widgets.validated_widget import ValidatedWidget
from gm_qt_plugin.exception_wrapper import ExceptionWrapper


class SignInWidget(QtWidgets.QWidget, QtPluginAccessor):
    WINDOW_WIDTH = 340
    WINDOW_HEIGHT = 493
    WINDOW_MARGINS = (0, 0, 0, 0)

    MARGIN_LEFT = 20
    MARGIN_TOP = 15
    MARGIN_RIGHT = 20
    MARGIN_BOTTOM = 20

    CONTENT_WIDTH = WINDOW_WIDTH - (MARGIN_LEFT + MARGIN_RIGHT)

    SIGN_IN_TEXT = u"Sign-in"
    SIGNING_IN_TEXT = u"Signing in..."

    SIGN_IN_ENVOY_TEXT = u"Sign-in with Envoy"
    SIGNING_IN_ENVOY_TEXT = u"Signing in with Envoy..."

    SIGN_IN_USING_ENVOY = u"Sign-in using Envoy"
    SIGN_IN_USING_CREDENTIALS = u"Sign-in using your Email and Access Key"

    CLOSE_BUTTON_WIDTH = 20
    PLUGIN_TITLE_IMAGE_WIDTH = 200

    def __init__(self, plugin, parent=None):
        QtWidgets.QWidget.__init__(self, parent=parent)
        QtPluginAccessor.__init__(self, plugin)

        # used for storing the position of the mouse for custom move events
        self._current_mouse_position = None

        """ Create Widgets """
        # title image
        self.title_image = QtWidgets.QLabel()
        self.title_image.setPixmap(qt_constants.GRIDMARKETS_IMAGE.scaledToWidth(150, QtCore.Qt.SmoothTransformation))
        self.title_image.setAlignment(QtCore.Qt.AlignCenter)

        # close window button
        self.close_window_btn = PushButton(u"X")
        self.close_window_btn.setFlat(True)
        self.close_window_btn.setFixedWidth(SignInWidget.CLOSE_BUTTON_WIDTH)
        self.close_window_btn.clicked.connect(self.close_window_slot)

        # The plugin specific title image
        plugin_title_image = plugin.get_user_interface().get_plugin_title_image()
        plugin_title_pixmap = plugin_title_image.scaledToWidth(SignInWidget.PLUGIN_TITLE_IMAGE_WIDTH,
                                                               QtCore.Qt.SmoothTransformation)
        self.plugin_title_label = QtWidgets.QLabel()
        self.plugin_title_label.setPixmap(plugin_title_pixmap)
        self.plugin_title_label.setAlignment(QtCore.Qt.AlignCenter)

        # the loading bar shown when the user is signing in
        self.loading_bar_movie = QtGui.QMovie(qt_constants.LOADING_BAR_GIF, parent=self)
        self.loading_bar_movie.setScaledSize(QtCore.QSize(75, 20))
        self.loading_bar_movie.setCacheMode(QtGui.QMovie.CacheAll)
        self.loading_bar = QtWidgets.QLabel("")
        self.loading_bar.setMovie(self.loading_bar_movie)

        self.sign_in_switch_widget = QtWidgets.QStackedWidget()

        ''' Standard Sign in screen widgets'''
        self.standard_sign_in_widget = QtWidgets.QWidget()

        # email input
        self.auth_email_input = ValidatedWidget(LineEdit())
        self.auth_email_input.get_widget().setPlaceholderText(u"GridMarkets Email")

        # access key input
        self.auth_access_key_input = ValidatedWidget(LineEdit())
        self.auth_access_key_input.get_widget().setPlaceholderText(u"GridMarkets Access Key")
        self.auth_access_key_input.get_widget().setEchoMode(QtWidgets.QLineEdit.Password)

        # sign in button
        self.sign_in_btn = ValidatedWidget(PushButton(SignInWidget.SIGN_IN_TEXT))
        self.sign_in_btn.align_warning_message_center(True)
        self.sign_in_btn.get_widget().clicked.connect(self.sign_in_slot)

        # lost your password button
        self.lost_password_link = PushButton(u"Lost your password?")
        self.lost_password_link.setFlat(True)
        self.lost_password_link.clicked.connect(
            lambda: qt_utils.open_url(qt_constants.FORGOT_PASSWORD_URL))

        ''' Envoy sign in screen widgets'''
        self.envoy_sign_in_widget = QtWidgets.QWidget()
        self.sign_in_envoy_btn = ValidatedWidget(PushButton(self.SIGN_IN_ENVOY_TEXT))
        self.sign_in_envoy_btn.align_warning_message_center(True)
        self.sign_in_envoy_btn.get_widget().clicked.connect(self.sign_in_envoy_slot)
        self.sign_in_envoy_btn.get_widget().setFocus()
        self.sign_in_envoy_btn.get_widget().setIcon(qt_constants.GM_ICON_WHITE)

        self.use_credentials_btn = PushButton(self.SIGN_IN_USING_CREDENTIALS)
        self.use_credentials_btn.setFlat(True)
        self.use_credentials_btn.clicked.connect(self.switch_sign_in_method_slot)

        self.use_envoy_btn = PushButton(self.SIGN_IN_USING_ENVOY)
        self.use_envoy_btn.setFlat(True)
        self.use_envoy_btn.clicked.connect(self.switch_sign_in_method_slot)

        # register
        self.register_label = QtWidgets.QLabel(u"Don't have an account?")
        self.register_link = PushButton(u"Create One")
        self.register_link.setFlat(True)
        self.register_link.clicked.connect(lambda: qt_utils.open_url(qt_constants.REGISTER_ACCOUNT_LINK))
        self.register_link.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)

        """ Create Layouts """
        self.layout = QtWidgets.QVBoxLayout(self)
        self.layout.setContentsMargins(self.MARGIN_LEFT, self.MARGIN_TOP, self.MARGIN_RIGHT, self.MARGIN_BOTTOM)

        self.loading_bar_layout = QtWidgets.QHBoxLayout(self)
        self.loading_bar_layout.setAlignment(QtCore.Qt.AlignHCenter)
        self.loading_bar_layout.addWidget(self.loading_bar)

        self.standard_sign_in_layout = QtWidgets.QVBoxLayout()
        self.standard_sign_in_layout.setContentsMargins(0, 0, 0, 0)
        self.standard_sign_in_layout.setAlignment(QtCore.Qt.AlignVCenter)
        self.standard_sign_in_layout.addWidget(self.auth_email_input)
        self.standard_sign_in_layout.addWidget(self.auth_access_key_input)
        self.standard_sign_in_layout.addSpacing(8)
        self.standard_sign_in_layout.addWidget(self.sign_in_btn)
        self.standard_sign_in_layout.addWidget(self.lost_password_link)
        self.standard_sign_in_layout.addWidget(self.use_envoy_btn)
        self.standard_sign_in_widget.setLayout(self.standard_sign_in_layout)

        self.envoy_sign_in_layout = QtWidgets.QVBoxLayout()
        self.envoy_sign_in_layout.setContentsMargins(0, 0, 0, 0)
        self.envoy_sign_in_layout.setAlignment(QtCore.Qt.AlignVCenter)
        self.envoy_sign_in_layout.addWidget(self.sign_in_envoy_btn)
        self.envoy_sign_in_layout.addWidget(self.use_credentials_btn)
        self.envoy_sign_in_widget.setLayout(self.envoy_sign_in_layout)

        self.sign_in_switch_widget.addWidget(self.envoy_sign_in_widget)
        self.sign_in_switch_widget.addWidget(self.standard_sign_in_widget)

        self.sign_in_container_layout = QtWidgets.QVBoxLayout()
        self.sign_in_container_layout.setAlignment(QtCore.Qt.AlignVCenter)
        self.sign_in_container_layout.addWidget(self.plugin_title_label)
        self.sign_in_container_layout.addWidget(self.sign_in_switch_widget)
        # self.sign_in_container_layout.addWidget(self.switch_sign_in_method_btn)

        self.register_layout = QtWidgets.QHBoxLayout()
        self.register_layout.setAlignment(QtCore.Qt.AlignHCenter)
        self.register_layout.setSpacing(0)
        self.register_layout.addWidget(self.register_label)
        self.register_layout.addWidget(self.register_link)

        self.top_corner_layout = QtWidgets.QVBoxLayout()
        self.top_corner_layout.setContentsMargins(0, 0, 0, 0)
        self.top_corner_layout.addWidget(self.close_window_btn)
        self.top_corner_layout.addSpacing(20)

        self.title_layout = QtWidgets.QHBoxLayout()
        self.title_layout.setAlignment(QtCore.Qt.AlignTop)
        self.title_layout.addSpacing(20)
        self.title_layout.addWidget(self.title_image)
        self.title_layout.addLayout(self.top_corner_layout)

        # self.layout.addWidget(self.title_image)
        self.layout.addLayout(self.title_layout)
        self.layout.addLayout(self.sign_in_container_layout)
        self.layout.addLayout(self.loading_bar_layout)
        self.layout.addLayout(self.register_layout)

    def get_title_image(self):
        return qt_constants.GRIDMARKETS_IMAGE.scaledToWidth(self.layout.width(), QtCore.Qt.SmoothTransformation)

    def _reset_input_validity(self):
        """ Resets all the error messages from the sign in window """

        user_interface = self.get_plugin().get_user_interface()
        user_interface.set_auth_email_validity_flag(True)
        user_interface.set_auth_access_key_validity_flag(True)
        user_interface.set_user_validity_flag(True)
        user_interface.set_auth_email_validity_message("")
        user_interface.set_auth_access_key_validity_message("")
        user_interface.set_user_validity_message("")

    def sign_in_slot(self):
        """ Sign in using the email and access key provided by the user """

        # get the email
        auth_email = self.auth_email_input.get_widget().text()

        # get the access key
        auth_access_key = self.auth_access_key_input.get_widget().text()

        # create a user instance with the above credentials
        user = gm_meta_plugin.User(auth_email, auth_access_key)

        self._worker_thread = WorkerThread(self.get_plugin())
        self._worker_thread.started.connect(lambda: self.show_status_message(u"Signing in..."))
        self._worker_thread.finished.connect(self.hide_status_message)

        # set all inputs to valid while signing in
        self._reset_input_validity()

        ################################################################################################################

        target = self.get_plugin().get_api_client().sign_in

        args = (
            user,
        )

        kwargs = {

        }

        self._worker_thread.process(target, args, kwargs, self.sign_in_result_handler)

    def sign_in_envoy_slot(self):
        self._worker_thread = WorkerThread(self.get_plugin())
        self._worker_thread.started.connect(lambda: self.show_status_message(u"Signing in..."))
        self._worker_thread.finished.connect(self.hide_status_message)

        # set all inputs to valid while signing in
        self._reset_input_validity()

        ################################################################################################################

        target = self.get_plugin().get_api_client().sign_in_with_envoy

        args = set()

        kwargs = {

        }

        self._worker_thread.process(target, args, kwargs, self.sign_in_result_handler)

    def sign_in_result_handler(self, result):
        """ Handel the result of the sign in operation """

        plugin = self.get_plugin()
        user_interface = plugin.get_user_interface()

        # None is returned when the sign in operation was successful
        if isinstance(result, QtNone):
            user_interface.set_auth_email_validity_flag(True)
            user_interface.set_auth_access_key_validity_flag(True)
            user_interface.set_user_validity_flag(True)
            user_interface.set_auth_email_validity_message("")
            user_interface.set_auth_access_key_validity_message("")
            user_interface.set_user_validity_message("")

            plugin.create_default_job_presets()

            main_window = self.get_plugin().get_user_interface().get_main_window()
            main_window.display_signed_in_ui()
            return True

        # handel expected exceptions
        if isinstance(result, ExceptionWrapper):
            exception = result.get_exception()

            if isinstance(exception, gm_meta_plugin.errors.InvalidEmailError):
                user_interface.set_auth_email_validity_flag(False)
                user_interface.set_auth_access_key_validity_flag(True)
                user_interface.set_user_validity_flag(True)
                user_interface.set_auth_email_validity_message(exception.user_message)
                user_interface.set_auth_access_key_validity_message("")
                user_interface.set_user_validity_message("")
                return True

            elif isinstance(exception, gm_meta_plugin.errors.InvalidAccessKeyError):
                user_interface.set_auth_email_validity_flag(True)
                user_interface.set_auth_access_key_validity_flag(False)
                user_interface.set_user_validity_flag(True)
                user_interface.set_auth_email_validity_message("")
                user_interface.set_auth_access_key_validity_message(exception.user_message)
                user_interface.set_user_validity_message("")
                return True

            elif isinstance(exception, gm_meta_plugin.errors.InvalidUserError):
                user_interface.set_auth_email_validity_flag(True)
                user_interface.set_auth_access_key_validity_flag(True)
                user_interface.set_user_validity_flag(False)
                user_interface.set_auth_email_validity_message("")
                user_interface.set_auth_access_key_validity_message("")
                user_interface.set_user_validity_message(exception.user_message)
                return True

            elif isinstance(exception, gm_meta_plugin.errors.APIClientError):
                user_interface.set_auth_email_validity_flag(True)
                user_interface.set_auth_access_key_validity_flag(True)
                user_interface.set_user_validity_flag(False)
                user_interface.set_auth_email_validity_message("")
                user_interface.set_auth_access_key_validity_message("")
                user_interface.set_user_validity_message(exception.user_message)
                return True

        return False

    def show_status_message(self, message):
        # play the loading gif
        self.loading_bar_movie.start()

        # disable the email input
        self.auth_email_input.get_widget().setEnabled(False)

        # disable the access key input
        self.auth_access_key_input.get_widget().setEnabled(False)

        # disable the sign in buttons
        self.sign_in_btn.get_widget().setEnabled(False)
        self.sign_in_envoy_btn.get_widget().setEnabled(False)

        # disable the sign in switch buttons
        self.use_credentials_btn.setEnabled(False)
        self.use_envoy_btn.setEnabled(False)

        # update the text in the sign in buttons to reflect what is going on
        self.sign_in_btn.get_widget().setText(SignInWidget.SIGNING_IN_TEXT)
        self.sign_in_envoy_btn.get_widget().setText(SignInWidget.SIGNING_IN_ENVOY_TEXT)

    def hide_status_message(self):
        # stop and rewind the loading gif
        self.loading_bar_movie.stop()
        self.loading_bar_movie.jumpToFrame(0)

        # enable the email input
        self.auth_email_input.get_widget().setEnabled(True)

        # enable the access key input
        self.auth_access_key_input.get_widget().setEnabled(True)

        # enable the sign in buttons
        self.sign_in_btn.get_widget().setEnabled(True)
        self.sign_in_envoy_btn.get_widget().setEnabled(True)

        # enable the sign in switch buttons
        self.use_credentials_btn.setEnabled(True)
        self.use_envoy_btn.setEnabled(True)

        # reset the sign in button's text labels
        self.sign_in_btn.get_widget().setText(SignInWidget.SIGN_IN_TEXT)
        self.sign_in_envoy_btn.get_widget().setText(SignInWidget.SIGN_IN_ENVOY_TEXT)

    def switch_sign_in_method_slot(self):
        """ Switch between 'sign in with Envoy' and 'sign in with credentials' """

        # animate opacity
        def fade_out_widget(widget):
            self.opacity_effect = QtWidgets.QGraphicsOpacityEffect(widget)
            widget.setGraphicsEffect(self.opacity_effect)
            self.a = QtCore.QPropertyAnimation(self.opacity_effect, "opacity")
            self.a.setDuration(400)
            self.a.setStartValue(1)
            self.a.setEndValue(0)
            self.a.setEasingCurve(QtCore.QEasingCurve.OutCubic)
            self.a.start(QtCore.QPropertyAnimation.DeleteWhenStopped)
            self.a.finished.connect(self.switch_sign_in_method_show_slot)

        if self.sign_in_switch_widget.currentWidget() == self.envoy_sign_in_widget:
            fade_out_widget(self.envoy_sign_in_widget)
        else:
            fade_out_widget(self.standard_sign_in_widget)

    def switch_sign_in_method_show_slot(self):

        # reset validation warnings
        self._reset_input_validity()

        # animate opacity
        def fade_in_widget(widget):
            self.opacity_effect = QtWidgets.QGraphicsOpacityEffect(widget)
            widget.setGraphicsEffect(self.opacity_effect)
            self.a = QtCore.QPropertyAnimation(self.opacity_effect, "opacity")
            self.a.setDuration(400)
            self.a.setStartValue(0)
            self.a.setEndValue(1)
            self.a.setEasingCurve(QtCore.QEasingCurve.InCubic)
            self.a.start(QtCore.QPropertyAnimation.DeleteWhenStopped)

        if self.sign_in_switch_widget.currentWidget() == self.envoy_sign_in_widget:
            fade_in_widget(self.standard_sign_in_widget)
            self.sign_in_switch_widget.setCurrentWidget(self.standard_sign_in_widget)
        else:
            fade_in_widget(self.envoy_sign_in_widget)
            self.sign_in_switch_widget.setCurrentWidget(self.envoy_sign_in_widget)

    def close_window_slot(self):
        """ Called when the user presses the close window button 'X' """
        main_window = self.get_plugin().get_user_interface().get_main_window()
        main_window.hide()

    def mousePressEvent(self, event):
        """ Cache the position of the mouse when first pressed """
        self._current_mouse_position = event.globalPos()

    def mouseMoveEvent(self, event):
        """ Move the window with the user's mouse when the user clicks and drags on any part of this widget """

        if self._current_mouse_position is not None:
            main_window = self.get_plugin().get_user_interface().get_main_window()

            # calculate the delta between the current position of the mouse and the mouses previous position
            delta = QtCore.QPoint(event.globalPos() - self._current_mouse_position)

            # move the window to be under the mouse
            main_window.move(main_window.x() + delta.x(), main_window.y() + delta.y())

            # cache the new position of the mouse
            self._current_mouse_position = event.globalPos()
