__all__ = 'MayaAttributeSource'

import sys
import re
import pathlib

import maya.cmds as cmds
import maya.mel

import gm_meta_plugin
from gm_meta_plugin.gridmarkets import constants as api_constants
from gm_meta_plugin.errors import ApplicationAttributeNotFound
from gm_meta_plugin.attribute_types import EnumItem
from gm_maya_plugin import maya_constants, maya_utils


def _get_machine_options(plugin, plugins=None, use_gpu_default=False):

    items = []

    api_client = plugin.get_api_client()
    machines = api_client.get_machines(api_constants.PRODUCTS.MAYA,
                                       api_constants.OPERATION_TYPES.RENDER,
                                       plugins=plugins)

    index = 0
    default = 0
    default_machines = []
    for machine in machines:
        items.append(EnumItem(machine.get_id(), machine.get_name(), u""))

        if machine.is_default():
            default_machines.append((machine, index))

        index = index + 1

    # if at least one machine was marked as default
    if default_machines:

        # use the first default machine who's is_gpu_machine() getter matches the use_gpu_default key word argument
        for default_machine in default_machines:
            if default_machine[0].is_gpu_machine() == use_gpu_default:
                default = default_machine[1]
                break
        else: # if no mach is found then use the first default machine in the list
            default = default_machines[0][1]

    return items, default


class ThirdPartyRendererAttributeSource(gm_meta_plugin.ApplicationAttributeSource):

    def __init__(self, plugin):
        gm_meta_plugin.ApplicationAttributeSource.__init__(self, plugin)

    @staticmethod
    def setup():
        raise NotImplementedError


class ArnoldAttributeSource(ThirdPartyRendererAttributeSource):
    """ A simple attribute source for mtoa specific settings """

    def __init__(self, plugin):
        ThirdPartyRendererAttributeSource.__init__(self, plugin)

    @staticmethod
    def setup():
        """
        The nodes holding Arnold's settings are created at runtime by mtoa. Before we access any of mtoa's settings
        they must be initialised by calling the mtoa.core.createOptions() method.
        """
        import mtoa.core as core
        core.createOptions()

    def get_enum_items(self, app, version, key):
        ArnoldAttributeSource.setup()

        if key == api_constants.API_KEYS.MACHINE_TYPE:
            return _get_machine_options(plugin=self.get_plugin(),
                                        plugins=[maya_constants.RENDERERS.MAYA_ARNOLD_RENDERER.get_commandline_name()],
                                        use_gpu_default=False)

        raise ApplicationAttributeNotFound(app, version, key)

    def get_attribute_value(self, app, version, key):
        ArnoldAttributeSource.setup()

        if key == api_constants.API_KEYS.RENDERER:
            return maya_constants.RENDERERS.MAYA_ARNOLD_RENDERER.get_commandline_name()

        elif key == api_constants.API_KEYS.OUTPUT_FORMAT:
            return maya.mel.eval("getAttr \"defaultArnoldDriver.aiTranslator\"")

        raise ApplicationAttributeNotFound(app, version, key)


class VrayAttributeSource(ThirdPartyRendererAttributeSource):
    """ A simple attribute source for vray specific settings """

    @staticmethod
    def setup():
        pass

    def get_enum_items(self, app, version, key):
        VrayAttributeSource.setup()

        if key == api_constants.API_KEYS.OUTPUT_FORMAT:

            # get all the menu items for the image format menu
            menu_items = maya.utils.executeInMainThreadWithResult(lambda: cmds.optionMenuGrp("vrayImageFormatMenu", query=True, itemListShort=True))

            # convert the menu items into EnumItems
            items = []
            for menu_item in menu_items:

                # if the menu_item is a divider then skip it
                if maya.utils.executeInMainThreadWithResult(lambda: cmds.menuItem(menu_item, query=True, divider=True)):
                    continue

                label = maya.utils.executeInMainThreadWithResult(lambda: cmds.menuItem(menu_item, query=True, label=True))

                # if the label contains a bracketed image format
                if MayaAttributeSource.bracketed_image_format_regex.match(label):
                    # extract the value within the brackets and use this as the value
                    start = label.index('(') + 1
                    end = label.index(')')
                    value = label[start:end]
                else:
                    value = label

                items.append(EnumItem(value, label, u""))

            return items, self._get_enum_index(app, version, key, items)

        if key == api_constants.API_KEYS.MACHINE_TYPE:
            return _get_machine_options(plugin=self.get_plugin(),
                                        plugins=[maya_constants.RENDERERS.MAYA_VRAY_RENDERER.get_commandline_name()],
                                        use_gpu_default=False)

        raise ApplicationAttributeNotFound(app, version, key)

    def get_attribute_value(self, app, version, key):
        VrayAttributeSource.setup()

        if key == api_constants.API_KEYS.OUTPUT_FORMAT:
            return maya.mel.eval("getAttr -as \"vraySettings.imageFormatStr\"")

        if key == api_constants.API_KEYS.RENDERER:
            return maya_constants.RENDERERS.MAYA_VRAY_RENDERER.get_commandline_name()

        raise ApplicationAttributeNotFound(app, version, key)


class RedshiftAttributeSource(ThirdPartyRendererAttributeSource):
    """ A simple attribute source for redshift specific settings """

    @staticmethod
    def setup():
        pass

    def get_redshift_image_format_labels(self):
        return maya.mel.eval("global string $g_redshiftImageFormatLabels[]; "
                             "$temp_var = $g_redshiftImageFormatLabels;")

    def get_redshift_image_format_values(self):
        return maya.mel.eval("global string $g_redshiftImageFormatLabels[]; "
                             "$temp_var = $g_redshiftImageFormatExtensions;")

    def get_enum_items(self, app, version, key):
        RedshiftAttributeSource.setup()

        if key == api_constants.API_KEYS.OUTPUT_FORMAT:

            items = []
            labels = self.get_redshift_image_format_labels()
            values = self.get_redshift_image_format_values()

            while labels:
                items.append(EnumItem(values.pop(0), labels.pop(0), u""))

            return items, self._get_enum_index(app, version, key, items)

        if key == api_constants.API_KEYS.MACHINE_TYPE:
            return _get_machine_options(plugin=self.get_plugin(),
                                        plugins=[maya_constants.RENDERERS.MAYA_REDSHIFT_RENDERER.get_commandline_name()],
                                        use_gpu_default=True)

        raise ApplicationAttributeNotFound(app, version, key)

    def get_attribute_value(self, app, version, key):
        RedshiftAttributeSource.setup()

        if key == api_constants.API_KEYS.OUTPUT_FORMAT:
            # redshift stores the image format as a long not an enum, so asString does not work
            index = maya.mel.eval("getAttr \"redshiftOptions.imageFormat\"")
            values = self.get_redshift_image_format_values()

            try:
                return values[index]
            except IndexError:
                return 'Unrecognised Image Format'

        if key == api_constants.API_KEYS.RENDERER:
            return maya_constants.RENDERERS.MAYA_REDSHIFT_RENDERER.get_commandline_name()

        raise ApplicationAttributeNotFound(app, version, key)


class RendermanAttributeSource(ThirdPartyRendererAttributeSource):
    """ A simple attribute source for rman specific settings """

    def __init__(self, plugin):
        ThirdPartyRendererAttributeSource.__init__(self, plugin)

    @staticmethod
    def setup():
        pass

    def get_enum_items(self, app, version, key):
        RendermanAttributeSource.setup()
        import rfm2

        if key == api_constants.API_KEYS.OUTPUT_FORMAT:
            rman_img_exts = rfm2.constants.IMG_EXTS
            items = list(map(lambda x: EnumItem(x, x, u""), rman_img_exts))
            return items, self._get_enum_index(app, version, key, items)

        if key == api_constants.API_KEYS.MACHINE_TYPE:
            return _get_machine_options(plugin=self.get_plugin(),
                                        plugins=[maya_constants.RENDERERS.MAYA_RENDERMAN_RENDERER.get_commandline_name()],
                                        use_gpu_default=False)

        raise ApplicationAttributeNotFound(app, version, key)

    def get_attribute_value(self, app, version, key):
        RendermanAttributeSource.setup()
        import rfm2

        if key == api_constants.API_KEYS.RENDERER:
            return maya_constants.RENDERERS.MAYA_RENDERMAN_RENDERER.get_commandline_name()

        elif key == api_constants.API_KEYS.OUTPUT_FORMAT:
            # see Pixar\RenderManForMaya-23.2\scripts\rfm2\api\strings.py set_display

            driver = None
            dspy_node = rfm2.config.DEFAULT_DISPLAY_NODE
            dspy_node = dspy_node.replace('_variance', '')
            if maya.utils.executeInMainThreadWithResult(lambda: cmds.objExists(dspy_node)):
                # connectionInfo() was failing: use listConnections() instead !
                try:
                    tmp = cmds.listConnections('%s.displayType' % dspy_node)
                except:
                    pass
                else:
                    driver = None if not tmp else tmp[0]
            if driver:
                ext = rfm2.config.cfg().displayDriverExtensions[maya.utils.executeInMainThreadWithResult(lambda: cmds.nodeType(driver))]
            else:
                ext = 'exr'

            return ext

        raise ApplicationAttributeNotFound(app, version, key)


class MayaAttributeSource(gm_meta_plugin.ApplicationAttributeSource):

    # A regular expression that detects image formats of the form
    # "<image format display name> (<image format command line name>)"
    # so that we can extract the value within the brackets
    bracketed_image_format_regex = re.compile("(.*\(.*\))")

    def __init__(self, plugin):
        gm_meta_plugin.ApplicationAttributeSource.__init__(self, plugin)

        MayaAttributeSource.setup()

        self._arnold_as = ArnoldAttributeSource(plugin)
        self._vray_as = VrayAttributeSource(plugin)
        self._redshift_as = RedshiftAttributeSource(plugin)
        self._renderman_as = RendermanAttributeSource(plugin)

        # create a dictionary of attribute sources for the different renderers.
        self._renderer_attribute_source_dict = {
            maya_constants.RENDERERS.MAYA_ARNOLD_RENDERER.get_internal_name(): self._arnold_as,
            maya_constants.RENDERERS.MAYA_VRAY_RENDERER.get_internal_name(): self._vray_as,
            maya_constants.RENDERERS.MAYA_REDSHIFT_RENDERER.get_internal_name(): self._redshift_as,
            maya_constants.RENDERERS.MAYA_RENDERMAN_RENDERER.get_internal_name(): self._renderman_as
        }

    @staticmethod
    def setup():
        """ Make sure the render settings window has been initialised so that values can be extracted from its menus"""

        if not maya_utils.render_settings_menu_exists():
            # open the maya renderer settings window to initialise it
            maya.mel.eval("unifiedRenderGlobalsWindow;")

            # process idle events to give the render settings window a chance to load. Not required but is good practice
            # Remember the plug-in may be set to auto-load so there will be alot of other deferred events at startup.
            maya_utils.process_idle_events(counter_limit=300)

            # hide the render settings window
            #maya.mel.eval("workspaceControl - e - vis 0 unifiedRenderGlobalsWindow;")

    def get_enum_items(self, app, version, key):

        current_renderer = maya_utils.get_current_renderer()

        # if the request can be fulfilled by one of the renderer attribute sources then use that attribute source
        # instead
        try:
            renderer_attribute_source = self._renderer_attribute_source_dict[current_renderer]

            try:
                return renderer_attribute_source.get_enum_items(app, version, key)
            except ApplicationAttributeNotFound as e:

                # if the renderer's attribute source exists but does not support this attribute key then call its setup
                # method so that its attributes are available
                renderer_attribute_source.setup()

        except KeyError as e:
            # if the renderer does not match a renderer attribute source in the attribute source dictionary then pass
            pass

        if key == api_constants.API_KEYS.OUTPUT_FORMAT:
            # get all the menu items for the image format menu
            menu_items = maya.utils.executeInMainThreadWithResult(lambda: cmds.optionMenuGrp("imageMenuMayaSW", query=True, itemListShort=True))

            # convert the menu items into EnumItems
            items = []
            for menu_item in menu_items:

                # if the menu_item is a divider then skip it
                if maya.utils.executeInMainThreadWithResult(lambda: cmds.menuItem(menu_item, query=True, divider=True)):
                    continue

                label = maya.utils.executeInMainThreadWithResult(lambda: cmds.menuItem(menu_item, query=True, label=True))

                # if the label contains a bracketed image format
                if MayaAttributeSource.bracketed_image_format_regex.match(label):
                    # extract the value within the brackets and use this as the value
                    start = label.index('(') + 1
                    end = label.index(')')
                    value = label[start:end]
                else:
                    value = label

                items.append(EnumItem(value, label, u""))

            return items, self._get_enum_index(app, version, key, items)

        if key == api_constants.API_KEYS.APP_VERSION:
            items = []

            current_maya_version = maya_utils.get_maya_major_version()
            items.append(EnumItem(current_maya_version, current_maya_version, u""))

            return items, 0

        if key == api_constants.API_KEYS.CAMERA:
            camera_options = maya.utils.executeInMainThreadWithResult(lambda: cmds.listCameras())
            items = list(map(lambda x: EnumItem(x, x, u""), camera_options))
            return items, self._get_enum_index(app, version, key, items)

        if key == api_constants.API_KEYS.MACHINE_TYPE:
            return _get_machine_options(plugin=self.get_plugin(), use_gpu_default=False)

        raise ApplicationAttributeNotFound(app, version, key)

    def get_attribute_value(self, app, app_version, key):

        current_renderer = maya_utils.get_current_renderer()

        # if the request can be fulfilled by one of the renderer attribute sources then use that attribute source
        # instead
        try:
            renderer_attribute_source = self._renderer_attribute_source_dict[current_renderer]
            return renderer_attribute_source.get_attribute_value(app, app_version, key)
        except KeyError as e:
            # if the renderer does not match a renderer attribute source in the attribute source dictionary then pass
            pass
        except ApplicationAttributeNotFound as e:
            # if the renderer's attribute source exists but does not support this attribute key pass
            pass

        if key == api_constants.API_KEYS.PROJECT_NAME:
            maya_project_directory = maya_utils.get_project_dir()
            return unicode(maya_project_directory.stem, sys.getfilesystemencoding())

        if key == api_constants.API_KEYS.JOB_NAME:
            return u"Maya Job (" + maya_utils.get_current_renderer() + u")"

        if key == api_constants.API_KEYS.APP:
            return api_constants.PRODUCTS.MAYA

        if key == api_constants.API_KEYS.APP_VERSION:
            return maya.utils.executeInMainThreadWithResult(lambda: cmds.about(version=True))

        if key == api_constants.API_KEYS.SCENE_FILE_PATH_UNUSED:
            file_name = maya.utils.executeInMainThreadWithResult(lambda: cmds.file(query=True, sceneName=True, shortName=True))
            return file_name

        if key == api_constants.API_KEYS.RENDERER:
            current_renderer = maya_utils.get_current_renderer()

            if current_renderer == maya_constants.RENDERERS.MAYA_SOFTWARE_RENDERER.get_internal_name():
                return maya_constants.RENDERERS.MAYA_SOFTWARE_RENDERER.get_commandline_name()

            elif current_renderer == maya_constants.RENDERERS.MAYA_HARDWARE_RENDERER.get_internal_name():
                return maya_constants.RENDERERS.MAYA_HARDWARE_RENDERER.get_commandline_name()

            elif current_renderer == maya_constants.RENDERERS.MAYA_HARDWARE_2_RENDERER.get_internal_name():
                return maya_constants.RENDERERS.MAYA_HARDWARE_2_RENDERER.get_commandline_name()

            elif current_renderer == maya_constants.RENDERERS.MAYA_VECTOR_RENDERER.get_internal_name():
                return maya_constants.RENDERERS.MAYA_VECTOR_RENDERER.get_commandline_name()

            elif current_renderer == maya_constants.RENDERERS.MAYA_RENDERMAN_RENDERER.get_internal_name():
                return maya_constants.RENDERERS.MAYA_RENDERMAN_RENDERER.get_commandline_name()

            else:
                return maya_constants.UNSUPPORTED_RENDERER + u" ({0})".format(current_renderer)

        if key == api_constants.API_KEYS.OUTPUT_FORMAT:
            image_format = maya.utils.executeInMainThreadWithResult(lambda: cmds.optionMenuGrp("imageMenuMayaSW", query=True, value=True))

            # if the label contains a bracketed image format
            if MayaAttributeSource.bracketed_image_format_regex.match(image_format):
                # extract the value within the brackets and use this as the value
                start = image_format.index('(') + 1
                end = image_format.index(')')
                value = image_format[start:end]
            else:
                value = image_format

            return value

        if key == api_constants.API_KEYS.FRAMES:
            start_frame = int(maya.utils.executeInMainThreadWithResult(lambda: cmds.getAttr("defaultRenderGlobals.startFrame")))
            end_frame = int(maya.utils.executeInMainThreadWithResult(lambda: cmds.getAttr("defaultRenderGlobals.endFrame")))
            frame_step = int(maya.utils.executeInMainThreadWithResult(lambda: cmds.getAttr("defaultRenderGlobals.byFrame")))
            return "{0} {1} {2}".format(start_frame, end_frame, frame_step)

        if key == api_constants.API_KEYS.CAMERA:
            # return the first camera in the list camera list
            # Maya by default always has at least four cameras, persp, top, front & side which can not be deleted
            return maya.utils.executeInMainThreadWithResult(lambda: cmds.listCameras()[0])

        if key == api_constants.API_KEYS.OUTPUT_WIDTH:
            return maya.utils.executeInMainThreadWithResult(lambda: cmds.getAttr("defaultResolution.width"))

        if key == api_constants.API_KEYS.OUTPUT_HEIGHT:
            return maya.utils.executeInMainThreadWithResult(lambda: cmds.getAttr("defaultResolution.height"))

        if key == api_constants.API_KEYS.LOCAL_DOWNLOAD_PATH:
            images_path = pathlib.Path(maya.utils.executeInMainThreadWithResult(lambda: cmds.workspace(fileRuleEntry="images")))
            root_dir = maya_utils.get_project_dir()
            local_output_path = (root_dir / images_path)
            return unicode(local_output_path)

        raise ApplicationAttributeNotFound(app, app_version, key)
