import maya.cmds

import pathlib

from PySide2 import QtWidgets
from maya import OpenMayaUI as omui
from shiboken2 import wrapInstance
from gm_qt_plugin import qt_constants
from gm_maya_plugin import maya_constants
import gm_meta_plugin


def maya_main_window():
    """ Helper method which returns the main Maya window """
    main_window_ptr = omui.MQtUtil.mainWindow()
    return wrapInstance(long(main_window_ptr), QtWidgets.QWidget)


def get_maya_scene_name():
    return maya.cmds.file(q=True, sceneName=True, shortName=True)


def does_plugin_menu_item_exist():
    return maya.cmds.menu("MayaWindow|GridMarkets", parent="MayaWindow", exists=True)


def add_actions_to_maya_menu_bar_item(main_plugin_window):
    def show_window(args):
        main_plugin_window.show()

    maya.cmds.menuItem(label='Open GridMarkets Plugin', image=qt_constants.GM_ICON_WHITE_PATH, command=show_window)


def get_maya_major_version():
    return maya.cmds.about(version=True)


def get_current_renderer():
    renderer = maya.cmds.getAttr("defaultRenderGlobals.currentRenderer")

    if renderer == 'renderman':
        renderer = 'rman'

    return renderer


def get_project_dir():
    return pathlib.Path(maya.cmds.workspace(query=True, rd=True))


def create_maya_menu_bar_btn(main_plugin_window):
    # If the menu bar item does not already exist create it
    if not does_plugin_menu_item_exist():
        def show_window(args):
            main_plugin_window.show()

        # create the menu
        maya.cmds.menu("GridMarkets",
                       label="GridMarkets",
                       tearOff=1,
                       parent='MayaWindow',
                       version=get_maya_major_version())

        # add the actions to the menu
        add_actions_to_maya_menu_bar_item(main_plugin_window)

        # highlight the new menu
        maya.cmds.whatsNewHighlight(highlightOn=True)


def update_maya_menu_bar_actions(main_plugin_window):
    if does_plugin_menu_item_exist():
        # remove all old items from the menu
        items = maya.cmds.menu("MayaWindow|GridMarkets", q=True, itemArray=True)
        for item in items:
            maya.cmds.deleteUI(item)

        # rebuild menu
        add_actions_to_maya_menu_bar_item(main_plugin_window)


def remove_maya_menu_bar_btn():
    # remove the menu bar item if it exists
    if does_plugin_menu_item_exist():
        maya.cmds.deleteUI("MayaWindow|GridMarkets")
        maya.cmds.refresh()


def render_settings_menu_exists():
    return maya.cmds.optionMenuGrp("imageMenuMayaSW", exists=True)


def process_idle_events(counter_limit=20):
    # This list contains all the current events/commands in Maya's idle queue
    queued_cmds = maya.cmds.evalDeferred(list=True)

    # The counter here is for safety
    counter = 0

    # Idle events might spawn more idle events so we need to loop until the list is empty
    while queued_cmds:

        # if a counter limit is provided and counter has reached the limit then break
        if counter_limit is not None and counter > counter_limit:
            break

        maya.utils.processIdleEvents()
        queued_cmds = maya.cmds.evalDeferred(list=True)
        counter = counter + 1

